#include <sys/wait.h>

#include <err.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>

#include "test_utils.h"

/*
 * This program is used to test TCP UTO scenarios.
 *
 * The server and the client are simulated by forking the process. Both the
 * client and the server can send or accept UTO. After the connection is
 * established the server artifically breaks the link between them and tries
 * to send data. After a period of "downtime" the link is restored.
 *
 * The exit status is determined by the status of the connection after the
 * artificial downtime.
 */
int
main(int argc, char *argv[])
{
	int status;
	char *client_addr, *server_addr, *port;
	int client_snd, client_rcv, server_snd, server_rcv;
	int downtime, waittime;

	pid_t server;
	pid_t client;

	if (argc != 9)
		errx(EX_USAGE, "Usage: %s client_addr server_addr port"
		    " client_snd client_rcv server_snd server_rcv downtime",
		    argv[0]);

	client_addr = argv[1];
	server_addr = argv[2];
	port = argv[3];
	
	/* No error checking since every positive value is allowed. */
	client_snd = strtonum(argv[4], 0, INT_MAX, NULL);
	client_rcv = strtonum(argv[5], 0, INT_MAX, NULL);
	server_snd = strtonum(argv[6], 0, INT_MAX, NULL);
	server_rcv = strtonum(argv[7], 0, INT_MAX, NULL);
	downtime = strtonum(argv[8], 0, INT_MAX, NULL);

	waittime = (downtime) ? downtime + 120 : 0;


	if(getuid())
		errx(EX_USAGE, "Only root may run this test");

	server = fork();
	switch(server) {
	case -1:
		err(EX_OSERR, "fork");
		break;
	case 0:
		server_disconnect(server_addr, port, downtime, server_snd,
		    server_rcv);
		exit(EX_OK);
		break;
	}

	/* Give the server a chance to run. */
	sleep(1);

	client = fork();
	switch(client) {
	case -1:
		err(EX_OSERR, "fork");
		break;
	case 0:
		client_generic(client_addr, server_addr, port, client_snd,
		    client_rcv);
		exit(EX_OK);
		break;
	}

	sleep(waittime);

	/* Wait server. It shouldn't hang. */
	if (waitpid(server, &status, WNOHANG) <= 0)
		/* Server hangs. */
		exit(-1);

	/* 
	 * Return the return code of the server. The caller will decide if
	 * this means the test passed or failed.
	 */
	if (WIFEXITED(status))
		exit(WEXITSTATUS(status));

	return -1;
}
