#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <unistd.h>

#include "test_utils.h"


#define SERVER_UTO	1600
#define CLIENT_UTO	1200
#define WAITTIME	5


/*
 * Waits for connection and tests received UTO.
 * Send new UTO and tests again the received UTO.
 */
static void
server(char *addr, char *port)
{
	int listen_sock;
	int sock;
	static char buf[6000];
	ssize_t bytes;

	int send_uto, recv_uto;

	/* Bind and start listening. */
	listen_sock = listen_uto(addr, port, SERVER_UTO, 1);

	sock = accept(listen_sock, NULL, NULL);
	if (sock < 0)
		err(EX_UNAVAILABLE, "accept");

	/* See what the client sent. */
	uto_get(sock, &send_uto, &recv_uto);
	if (send_uto != SERVER_UTO || recv_uto != CLIENT_UTO)
		exit(-1);

	/* Change value and send data to let the client know. */
	sleep(2);
	uto_set(sock, 2 * SERVER_UTO, 1);

	bytes = send(sock, &buf, sizeof(buf), 0);
	if (bytes < 0)
		err(EX_IOERR, "send");

	/* The client should have sent us a new value. */
	bytes = recv(sock, &buf, sizeof(buf), MSG_WAITALL);
	if (bytes < 0 || (unsigned) bytes < sizeof(buf))
		err(EX_IOERR, "send");

	/* Fetch the new value and print it. */
	uto_get(sock, &send_uto, &recv_uto);
	if (send_uto != 2 * SERVER_UTO || recv_uto != 2 * CLIENT_UTO)
		exit(-1);
}

/*
 * Connects to the server and tests received UTO.
 * Exchanges data and tests the received value again.
 */
static void
client(char *addr, char *port)
{
	int sock;
	static char buf[6000];
	ssize_t bytes;
	int send_uto, recv_uto;

	/* Connect to the server, with the specified UTO values. */
	sock = connect_uto(addr, addr, port, CLIENT_UTO, 1);

	/* See what the server sent. */
	uto_get(sock, &send_uto, &recv_uto);
	if (send_uto != CLIENT_UTO || recv_uto != SERVER_UTO)
		exit(-1);

	/* Receive some data to make sure the server had a chance to send us
	 * his new idea of UTO */
	bytes = recv(sock, &buf, sizeof(buf), MSG_WAITALL);
	if (bytes < 0 || (unsigned) bytes < sizeof(buf))
		err(EX_IOERR, "send");

	/* Change value halfway. */
	uto_set(sock, 2 * send_uto, 1);

	/* Send dummy data to let the server know the new value. */
	bytes = send(sock, &buf, sizeof(buf), 0);
	if (bytes < 0)
		err(EX_IOERR, "send");

	uto_get(sock, &send_uto, &recv_uto);
	if (send_uto != 2 * CLIENT_UTO || recv_uto != 2 * SERVER_UTO)
		exit(-1);
}

/*
 * This test exchanges dummy data via TCP and checks that the expected
 * UTO values are received.
 *
 * The test passes if the children processes exit normally.
 */

int
main(int argc, char *argv[])
{
	int status;
	char *addr, *port;

	if (argc != 3)
		errx(EX_USAGE, "Usage: %s address port", argv[0]);

	addr = argv[1];
	port = argv[2];

	switch(fork()) {
	case -1:
		err(EX_OSERR, "fork");
		break;
	case 0:
		server(addr, port);
		exit(EX_OK);
		break;
	}

	/* Give the server a chance to run. */
	sleep(1);

	switch(fork()) {
	case -1:
		err(EX_OSERR, "fork");
		break;
	case 0:
		client(addr, port);
		exit(EX_OK);
		break;
	}
	sleep(WAITTIME);

	/* Check if both children terminated normally. */
	if (waitpid(-1, &status, WNOHANG) < 0 || !WIFEXITED(status) ||
	    WEXITSTATUS(status))
		/* Hangs or crashed. */
		exit(-1);

	if (waitpid(-1, &status, WNOHANG) < 0 || !WIFEXITED(status) ||
	    WEXITSTATUS(status))
		/* Hangs or crashed. */
		exit(-1);

	return EX_OK;
}
